// -------------------------------------------Please DO NOT change below---------------------------------------- //
import { httpRequest } from "http-request";
import { Cookies, SetCookie } from "cookies";

const DEFAULT_COOKIE_NAME = "datadome";
const DEFAULT_REQUEST_PORT = 0;

// If a value is not found, an empty string is retrieved instead.
function getHeaderValue(httpMessage, key) {
  if (httpMessage == null || key == null) {
    return "";
  }

  var val = httpMessage.getHeader(key);

  if (val != null && val.length > 0) {
    return val[0];
  }

  return "";
}

// Cleans and encodes the request body.
// N.B. Empty headers are removed.
function serializeQuery(obj) {
  var str = [];
  for (var p in obj) {
    var value = obj[p];
    value = value == null ? "" : value.toString();
    value = value.trim();

    if (value.length > 0) {
      str.push(encodeURIComponent(p) + "=" + encodeURIComponent(obj[p]));
    }
  }
  return str.join("&");
}

// Parses the DataDome cookie content and retrieves the clientId (if any), and cookies length.
function getClientIdAndCookiesLength(request) {
  var cookieHeader = request.getHeader("Cookie");
  var cookies = new Cookies(cookieHeader);

  var clientId = null;
  if (cookies !== null) {
    clientId = cookies.get(DEFAULT_COOKIE_NAME);
  }

  var cookiesLength = 0;
  if (cookieHeader != null && cookieHeader.length > 0) {
    cookiesLength = cookieHeader[0].length;
  }

  return {
    clientId: clientId,
    cookiesLength: cookiesLength,
  };
}

// Header truncation as required by DataDome: https://docs.datadome.co/reference/validate-request#api-field-size
function truncateHeaders(obj) {
  var limits = JSON.parse(
    '{"key":"","ip":"","port":"","protocol":"","method":"","cookieslen":"","authorizationlen":"","moduleversion":"","postparamlen":"","apiconnectionstate":"","requestmodulename":"","timerequest":"","secfetchuser":8,"secchdevicememory":8,"secchuamobile":8,"tlsprotocol":8,"secchuaarch":16,"contenttype":64,"secchuaplatform":32,"secfetchdest":32,"secfetchmode":32,"secfetchsite":64,"tlscipher":64,"clientid":128,"from":128,"x-requested-with":128,"acceptcharset":128,"acceptencoding":128,"connection":128,"pragma":128,"cachecontrol":128,"secchua":128,"secchuamodel":128,"trueclientip":128,"secchuafullversionlist":256,"acceptlanguage":256,"via":256,"headerslist":512,"origin":512,"serverhostname":512,"servername":512,"xforwardedforip":512,"accept":512,"host":512,"useragent":768,"referer":1024,"request":2048}'
  );

  var keys = Object.keys(obj);
  for (var i = 0; i < keys.length; i++) {
    var key = keys[i];
    var lowerKey = key.toLowerCase();
    var limit = limits[lowerKey];
    limit = limit == null ? "" : limit;

    var objVal = obj[key];
    if (objVal == null) {
      continue;
    }

    if (lowerKey == "xforwardedforip" && limit != "" && objVal.length > limit) {
      obj[key] = objVal.slice(-limit);
    } else if (limit != "" && objVal.length > limit) {
      obj[key] = objVal.substring(0, limit);
    }
  }

  return obj;
}

// Gets the request timestamp
function getCurrentMicroTime() {
  return new Date().getTime() * 1000;
}

// DataDome response processing
async function processDataDomeResponse(request, response) {
  // Only checking for status code 200 due to "X-DataDome-OK-On-Blocking"
  if (!(response.status == 200)) {
    return;
  }

  // Retrieving the status code from the DataDome response header.
  var strStatusCode = getHeaderValue(response, "X-DataDomeResponse");
  var statusCode = parseInt(strStatusCode, 10);
  if (isNaN(statusCode)) {
    // The "X-DataDomeResponse" header is crucial to determine whether this request should be allowed or blocked
    // If this value is not returned by DataDome, then we will allow this request to proceed to the origin
    return;
  }

  // Holds data to be retrieved in the X-DataDome-Log
  var debugHeaders = [];

  // Adding DataDome headers to the request - to reach the origin
  var strRequestHeaders = getHeaderValue(
    response,
    "X-DataDome-request-headers"
  );
  debugHeaders.push("DataDome Request Headers: " + strRequestHeaders);

  var requestHeaders = strRequestHeaders.split(" ");
  if (requestHeaders != null && requestHeaders.length > 0) {
    requestHeaders.forEach(function(rh) {
      var hv = getHeaderValue(response, rh);
      if (hv !== "") {
        request.setHeader(rh, hv);
      }
    });
  }

  // Retrieving DataDome headers to be later added to the response - to reach the user agent
  var strResponseHeaders = getHeaderValue(response, "X-DataDome-headers");
  debugHeaders.push("DataDome Response Headers: " + strResponseHeaders);

  var responseHeaders = strResponseHeaders.split(" ");

  var newHeaderObj = {};
  responseHeaders.forEach(function(rh) {
    var hv = getHeaderValue(response, rh);
    if (hv !== "") {
      newHeaderObj[rh] = hv;
    }
  });

  // Storing the status code and response headers in a flow variable, to be accessed from the onClientResponse function
  var flowVariable = {
    statusCode: statusCode,
    responseHeaders: newHeaderObj,
  };

  request.setVariable("PMUSER_EW_DD_FLOW", JSON.stringify(flowVariable));

  try {
    request.setVariable("PMUSER_EW_DD_LOG_HEADER", debugHeaders.join(" - "));
  } catch (logError) {}

  // If DataDome considers the request to be malicious, then the captcha/blocked page is returned
  if (
    statusCode == 301 ||
    statusCode == 302 ||
    statusCode == 401 ||
    statusCode == 403
  ) {
    var textResponse = await response.text();
    request.respondWith(statusCode, newHeaderObj, textResponse);
  }
}

export async function onClientRequest(request) {
  try {
    // Retrieving body values
    var cookiesRes = getClientIdAndCookiesLength(request);

    var requestHeaders = request.getHeaders();
    var requestHeaderKeysArray = Object.keys(requestHeaders);

    var body = {
      Key: request.getVariable("PMUSER_EW_DD_LICENSE_KEY"),
      IP: request.getVariable("PMUSER_EW_CLIENT_IP"),
      Accept: getHeaderValue(request, "Accept"),
      AcceptCharset: getHeaderValue(request, "Accept-Charset"),
      AcceptEncoding: getHeaderValue(request, "Accept-Encoding"),
      AcceptLanguage: getHeaderValue(request, "Accept-Language"),
      AuthorizationLen: getHeaderValue(request, "Authorization").length,
      CacheControl: getHeaderValue(request, "Cache-Control"),
      Connection: getHeaderValue(request, "Connection"),
      ContentType: getHeaderValue(request, "Content-Type"),
      CookiesLen: cookiesRes.cookiesLength,
      From: getHeaderValue(request, "From"),
      HeadersList: requestHeaderKeysArray.join(","),
      Host: request.host,
      Method: request.method,
      ModuleVersion: "1.2.0",
      Origin: getHeaderValue(request, "Origin"),
      Port: DEFAULT_REQUEST_PORT,
      PostParamLen: getHeaderValue(request, "Content-Length"),
      Pragma: getHeaderValue(request, "Pragma"),
      Protocol: request.scheme,
      Referer: getHeaderValue(request, "Referer"),
      Request: request.url,
      RequestModuleName: "akamai",
      SecCHDeviceMemory: getHeaderValue(request, "Sec-CH-Device-Memory"),
      SecCHUA: getHeaderValue(request, "Sec-CH-UA"),
      SecCHUAArch: getHeaderValue(request, "Sec-CH-UA-Arch"),
      SecCHUAFullVersionList: getHeaderValue(
        request,
        "Sec-CH-UA-Full-Version-List"
      ),
      SecCHUAMobile: getHeaderValue(request, "Sec-CH-UA-Mobile"),
      SecCHUAModel: getHeaderValue(request, "Sec-CH-UA-Model"),
      SecCHUAPlatform: getHeaderValue(request, "Sec-CH-UA-Platform"),
      SecFetchDest: getHeaderValue(request, "Sec-Fetch-Dest"),
      SecFetchMode: getHeaderValue(request, "Sec-Fetch-Mode"),
      SecFetchSite: getHeaderValue(request, "Sec-Fetch-Site"),
      SecFetchUser: getHeaderValue(request, "Sec-Fetch-User"),
      ServerHostname: request.host,
      ServerName: request.host,
      ServerRegion: request.userLocation?.region ?? "",
      TimeRequest: getCurrentMicroTime(),
      TlsCipher: request.getVariable("PMUSER_EW_TLS_CIPHER_NAME"),
      TlsProtocol: request.getVariable("PMUSER_EW_TLS_VERSION"),
      TrueClientIP: getHeaderValue(request, "True-Client-IP"),
      UserAgent: getHeaderValue(request, "User-Agent"),
      Via: getHeaderValue(request, "Via"),
      XForwardedForIP: getHeaderValue(request, "X-Forwarded-For"),
      "X-Real-IP": getHeaderValue(request, "X-Real-IP"),
      "X-Requested-With": getHeaderValue(request, "x-requested-with"),
    };

    // Due to the nature of the Akamai EWs, the call to DD will always return a 200 status code ("X-DataDome-OK-On-Blocking")
    // This was done to mitigate the keep-alive TLS handshake causing performance issues
    var headers = {
      "Content-Type": "application/x-www-form-urlencoded",
      "X-DataDome-OK-On-Blocking": "true",
    };

    // Retrieving the clientId value from either the header, or the cookie.
    var xDatadomeClientid = getHeaderValue(request, "x-datadome-clientid");
    if (
      xDatadomeClientid !== null &&
      xDatadomeClientid !== undefined &&
      xDatadomeClientid.length > 0
    ) {
      headers["x-datadome-x-set-cookie"] = "true";
      body["ClientId"] = xDatadomeClientid;
    } else {
      body["ClientId"] = cookiesRes.clientId;
    }

    body = truncateHeaders(body);

    const options = {
      method: "POST",
      body: serializeQuery(body),
      headers: headers,
    };

    options.headers["Content-Length"] = options.body.length;

    var strTimeout = request.getVariable("PMUSER_EW_DD_TIMEOUT");
    if (strTimeout == null) {
      // Defaulting to 100ms
      options["timeout"] = 100;
    } else {
      try {
        // Attempting to set the DataDome timeout parameter
        var timeout = Number.parseInt(strTimeout);
        options["timeout"] = timeout;
      } catch (timeoutError) {}
    }

    var apiURL = "/validate-request";
    var response = await httpRequest(apiURL, options);
    await processDataDomeResponse(request, response);
  } catch (error) {
    try {
      request.setVariable("PMUSER_EW_DD_LOG_HEADER", "Error: " + error);
    } catch (logError) {}
  }
}

export function onClientResponse(request, response) {
  // Retrieving the flow variable
  var strFlowVariable = request.getVariable("PMUSER_EW_DD_FLOW");

  if (strFlowVariable != null) {
    var flag = true;
    var flowVariable = {};

    try {
      flowVariable = JSON.parse(strFlowVariable);
    } catch (error) {
      flag = false;
    }

    // If the flow variable is valid, then ..
    if (flag) {
      // If the origin has been reached successfully, add the DataDome headers to the response
      if (flowVariable.statusCode == 200) {
        var keys = Object.keys(flowVariable.responseHeaders);
        for (var i = 0; i < keys.length; i++) {
          var key = keys[i];

          response.addHeader(key, flowVariable.responseHeaders[key]);
        }
      } else {
        // Injecting the Set-Cookie header during the response method as the "respondWith" does not support it
        var strCookie = flowVariable.responseHeaders["Set-Cookie"];
        if (strCookie != null && strCookie.length > 0) {
          var cookie = new SetCookie(strCookie);
          response.setHeader("Set-Cookie", cookie.toHeader());
        }
      }
    }
  }

  // Retrieving the debug variable
  var debugVariable = request.getVariable("PMUSER_EW_DD_ENABLE_LOGGING");
  if (debugVariable != null) {
    var includeDebugHeader = debugVariable === "true";

    // Injecting the DataDome Log header if the Debug variable is enabled
    if (includeDebugHeader) {
      response.addHeader(
        "X-DataDome-Log",
        request.getVariable("PMUSER_EW_DD_LOG_HEADER")
      );
    }
  }
}
